#
# EasyML expression writer
#
# Supported functions:
#  https://opencarp.org/documentation/examples/01_ep_single_cell/05_easyml
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import warnings

import myokit

from myokit.formats.ansic import CBasedExpressionWriter


class EasyMLExpressionWriter(CBasedExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` writes
    equations for variables in EasyML syntax.

    EasyML has a C-like syntax, and uses C operator precedence.
    """
    def __init__(self):
        super().__init__()

    #def _ex_name(self, e):
    #def _ex_derivative(self, e):
    #def _ex_number(self, e):
    #def _ex_prefix_plus(self, e):
    #def _ex_prefix_minus(self, e):
    #def _ex_plus(self, e):

    def _ex_minus(self, e):
        if isinstance(e[0], myokit.Exp) and isinstance(e[1], myokit.Number):
            if e[1].eval() == 1:
                return f'expm1({self.ex(e[0][0])})'
        if isinstance(e[1], myokit.Exp) and isinstance(e[0], myokit.Number):
            if e[0].eval() == 1:
                return f'-expm1({self.ex(e[1][0])})'
        return super()._ex_minus(e)

    #def _ex_multiply(self, e):
    #def _ex_divide(self, e):
    #def _ex_quotient(self, e):
    #def _ex_remainder(self, e):
    #def _ex_power(self, e):
    #def _ex_sqrt(self, e):

    def _ex_sin(self, e):
        warnings.warn('Unsupported function: sin()')
        return super()._ex_sin(e)

    #def _ex_cos(self, e):

    def _ex_tan(self, e):
        warnings.warn('Unsupported function: tan()')
        return super()._ex_tan(e)

    def _ex_asin(self, e):
        warnings.warn('Unsupported function: asin()')
        return super()._ex_asin(e)

    #def _ex_acos(self, e):

    def _ex_atan(self, e):
        warnings.warn('Unsupported function: atan()')
        return super()._ex_atan(e)

    #def _ex_exp(self, e):
    #def _ex_log(self, e):
    #def _ex_log10(self, e):

    def _ex_floor(self, e):
        warnings.warn('Unsupported function: floor()')
        return super()._ex_floor(e)

    def _ex_ceil(self, e):
        warnings.warn('Unsupported function: ceil()')
        return super()._ex_ceil(e)

    #def _ex_abs(self, e):

    #def _ex_equal(self, e):
    #def _ex_not_equal(self, e):
    #def _ex_more(self, e):
    #def _ex_less(self, e):
    #def _ex_more_equal(self, e):
    #def _ex_less_equal(self, e):

    #def _ex_not(self, e):
    #def _ex_and(self, e):
    #def _ex_or(self, e):

    #def _ex_if(self, e):
    #def _ex_piecewise(self, e):




#!/usr/bin/env python3
#
# Tests Myokit's SBML support.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import unittest

import myokit
import myokit.formats

import myokit.formats.sbml
from myokit.formats.sbml._api import Model
from myokit.formats.sbml._writer import write_string
import myokit.formats.sbml as sbml


class TestSBMLExport(unittest.TestCase):
    """
    Unit tests for the SBML export functionality.
    """

    def test_empty_model(self):
        # Test exporting an empty model
        model = Model()
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<sbml", sbml_str)
        self.assertIn("</sbml>", sbml_str)
        self.assertIn('<model id="unnamed_model"/>', sbml_str)

    def test_time_unit(self):
        # Test setting the time unit
        model = Model()
        model.set_time_units(myokit.units.second)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('timeUnits="second"', sbml_str)

        model = Model()
        model.set_time_units(1e3 * myokit.units.second)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('<listOfUnitDefinitions>', sbml_str)
        self.assertIn('<unitDefinition id="s_times_1e3">', sbml_str)

    def test_area_unit(self):
        # Test setting the area unit
        model = Model()
        model.set_area_units(myokit.units.metre)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('areaUnits="metre"', sbml_str)

    def test_volume_unit(self):
        # Test setting the volume unit
        model = Model()
        model.set_volume_units(myokit.units.litre)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('volumeUnits="litre"', sbml_str)

    def test_substance_unit(self):
        # Test setting the substance unit
        model = Model()
        model.set_substance_units(myokit.units.mole)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('substanceUnits="mole"', sbml_str)

    def test_extent_unit(self):
        # Test setting the extent unit
        model = Model()
        model.set_extent_units(myokit.units.mole)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('extentUnits="mole"', sbml_str)

    def test_length_unit(self):
        # Test setting the length unit
        model = Model()
        model.set_length_units(myokit.units.metre)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn('lengthUnits="metre"', sbml_str)

    def test_list_of_unit_definitions(self):
        # Test setting a list of unit definitions
        model = Model()
        model.add_unit("my_unit", myokit.units.ampere)
        model.add_unit("my_unit2", 2 * myokit.units.dimensionless)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfUnitDefinitions>", sbml_str)
        self.assertIn('<unitDefinition id="my_unit">', sbml_str)
        self.assertIn(
            '<unit kind="ampere" exponent="1.0" multiplier="1.0"/>',
            sbml_str
        )
        self.assertIn('<unitDefinition id="my_unit2">', sbml_str)
        self.assertIn(
            '<unit kind="dimensionless" multiplier="2.0"/>',
            sbml_str
        )

    def test_list_of_compartments(self):
        # Test setting a list of compartments
        model = Model()
        c = model.add_compartment("my_compartment")
        c.set_size_units(myokit.units.litre)
        c.set_spatial_dimensions(3)
        c = model.add_compartment("my_compartment2")
        c.set_size_units(1e3 * myokit.units.metre ** 3)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfCompartments>", sbml_str)
        self.assertIn(
            '<compartment id="my_compartment" units="litre" spatialDimensions="3.0"/>',  # noqa: E501
            sbml_str
        )
        self.assertIn(
            '<compartment id="my_compartment2" units="m3_times_1e3"/>',
            sbml_str
        )
        self.assertIn("<listOfUnitDefinitions>", sbml_str)
        self.assertIn('<unitDefinition id="m3_times_1e3">', sbml_str)

    def test_constants_and_literals(self):
        m = myokit.Model()
        c = m.add_component('comp')
        t = c.add_variable('time', rhs=myokit.Number(0))
        t.set_unit(myokit.units.second)
        t.set_binding('time')
        v = c.add_variable('my_parameter')
        v.set_rhs(myokit.Number(1))
        v2 = c.add_variable('my_parameter2')
        v2.set_rhs(myokit.Multiply(myokit.Number(2), myokit.Name(v)))

        # check that the equation is exported correctly to sbml.Model
        s = sbml.Model.from_myokit_model(m)
        parameter_names = [v.sid() for v in s.parameters()]
        self.assertCountEqual(
            parameter_names, ['my_parameter', 'my_parameter2']
        )
        sbml_v = s.parameter('my_parameter')
        sbml_v2 = s.parameter('my_parameter2')
        self.assertEqual(sbml_v.value(), myokit.Number(1))
        self.assertIsNone(sbml_v.initial_value())
        self.assertEqual(
            sbml_v2.initial_value(),
            myokit.Multiply(myokit.Number(2), myokit.Name(v))
        )
        self.assertIsNone(sbml_v2.value())
        self.assertTrue(sbml_v.is_constant())
        self.assertTrue(sbml_v2.is_constant())
        self.assertFalse(sbml_v2.is_literal())
        self.assertTrue(sbml_v.is_literal())

    def test_list_of_parameters(self):
        # Test setting a list of parameters
        model = Model()
        p = model.add_parameter("my_parameter")
        p.set_value(myokit.Number(1))
        p = model.add_parameter("my_parameter2")
        p.set_units(1e3 * myokit.units.metre / myokit.units.second)
        p.set_value(myokit.Number(2))
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfParameters>", sbml_str)
        self.assertIn(
            '<parameter id="my_parameter" constant="true" value="1.0"/>',
            sbml_str
        )
        self.assertIn(
            '<parameter id="my_parameter2" units="m_per_s_times_1e3" constant="true" value="2.0"/>',  # noqa: E501
            sbml_str
        )
        self.assertNotIn("<listOfRules>", sbml_str)
        self.assertNotIn("<listOfInitialAssignments>", sbml_str)
        self.assertIn("<listOfUnitDefinitions>", sbml_str)
        self.assertIn('<unitDefinition id="m_per_s_times_1e3">', sbml_str)

    def test_list_of_parameters_const_depend(self):
        # Test setting a constant parameter that depends on another parameter
        model = Model()
        p = model.add_parameter("my_parameter")
        p.set_value(myokit.Number(1))
        p = model.add_parameter("my_parameter2")
        p.set_initial_value(
            myokit.Multiply(myokit.Name("my_parameter"), myokit.Number(2))
        )
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfParameters>", sbml_str)
        self.assertIn(
            '<parameter id="my_parameter" constant="true" value="1.0"/>',
            sbml_str
        )
        self.assertIn(
            '<parameter id="my_parameter2" constant="true"/>',  # noqa: E501
            sbml_str
        )
        self.assertNotIn("<listOfRules>", sbml_str)
        self.assertIn("<listOfInitialAssignments>", sbml_str)
        self.assertIn('<initialAssignment symbol="my_parameter2">', sbml_str)
        self.assertIn("<times/>\n            <ci>my_parameter</ci>\n            <cn>2.0</cn>", sbml_str)  # noqa: E501

    def test_list_of_parameters_var(self):
        model = Model()
        p = model.add_parameter("my_parameter", is_constant=False)
        p.set_initial_value(myokit.Number(1))
        p.set_value(myokit.Number(2))
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfRules>", sbml_str)
        self.assertIn('<assignmentRule variable="my_parameter">', sbml_str)
        self.assertIn("<cn>2.0</cn>", sbml_str)
        self.assertIn("<listOfInitialAssignments>", sbml_str)
        self.assertIn('<initialAssignment symbol="my_parameter">', sbml_str)
        self.assertIn("<cn>1.0</cn>", sbml_str)

    def test_list_of_species(self):
        # Test setting a list of species
        model = Model()
        c = model.add_compartment("my_compartment")
        s = model.add_species(c, "my_species")
        s.set_substance_units(myokit.units.mole / myokit.units.litre)
        s.set_value(myokit.Number(1), True)
        s.set_initial_value(myokit.Number(2))
        s = model.add_species(c, "my_species2")
        s.set_substance_units(myokit.units.mole)
        s.set_value(myokit.Number(1), True)
        s.set_initial_value(myokit.Number(2), in_amount=True)
        s = model.add_species(c, "my_species3", is_amount=True)
        s.set_value(myokit.Number(1), True)
        s.set_initial_value(myokit.Number(2))
        s = model.add_species(c, "my_species4", is_amount=True)
        s.set_value(myokit.Number(1), True)
        s.set_initial_value(myokit.Number(2), in_amount=False)
        s = model.add_species(c, "my_species5")
        s.set_value(myokit.Number(1), False)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfSpecies>", sbml_str)
        self.assertIn(
            '<species id="my_species" compartment="my_compartment" initialConcentration="2.0" constant="False" units="M" boundaryCondition="False"/>',  # noqa: E501
            sbml_str,
        )
        self.assertIn(
            '<species id="my_species2" compartment="my_compartment" initialAmount="2.0" constant="False" units="mole" boundaryCondition="False"/>',  # noqa: E501
            sbml_str,
        )
        self.assertIn(
            '<species id="my_species3" compartment="my_compartment" initialAmount="2.0" constant="False" boundaryCondition="False"/>',  # noqa: E501
            sbml_str,
        )
        self.assertIn(
            '<species id="my_species4" compartment="my_compartment" initialConcentration="2.0" constant="False" boundaryCondition="False"/>',  # noqa: E501
            sbml_str,
        )
        self.assertIn("<listOfRules>", sbml_str)
        self.assertIn('<rateRule variable="my_species">', sbml_str)
        self.assertIn("<cn>1.0</cn>", sbml_str)
        self.assertIn('<assignmentRule variable="my_species5">', sbml_str)

    def test_list_of_reactions(self):
        # test kinetic law
        model = Model()
        c = model.add_compartment("my_compartment")
        s = model.add_species(c, "my_species")
        s2 = model.add_species(c, "my_species2")
        r = model.add_reaction("my_reaction")
        r.add_reactant(s, "my_reaction_reactant")
        r.add_product(s)
        r.add_modifier(s)
        r.add_modifier(s2, "my_modifier")
        r.set_kinetic_law(myokit.Number(1))
        sbml_str = write_string(model).decode("utf8")
        self.assertIn("<listOfReactions>", sbml_str)
        self.assertIn('<reaction id="my_reaction">', sbml_str)
        self.assertIn("<listOfReactants>", sbml_str)
        self.assertIn("<listOfProducts>", sbml_str)
        self.assertIn('<speciesReference species="my_species"/>', sbml_str)
        self.assertIn('<speciesReference species="my_species" id="my_reaction_reactant"/>', sbml_str)  # noqa: E501
        self.assertIn("<listOfModifiers>", sbml_str)
        self.assertIn(
            '<modifierSpeciesReference species="my_species"/>',
            sbml_str
        )
        self.assertIn(
            '<modifierSpeciesReference species="my_species2" id="my_modifier"/>',  # noqa: E501
            sbml_str
        )
        self.assertIn("<kineticLaw>", sbml_str)
        self.assertIn("<cn>1.0</cn>", sbml_str)

        # test stochoimetry
        model = Model()
        c = model.add_compartment("my_compartment")
        s = model.add_species(c, "my_species")
        r = model.add_reaction("my_reaction")
        react = r.add_reactant(s)
        react.set_value(myokit.Number(1))
        r.add_product(s)
        sbml_str = write_string(model).decode("utf8")
        self.assertIn(
            '<speciesReference species="my_species" stoichiometry="1.0"/>',
            sbml_str
        )

    def test_expressions_multiple_compartments(self):
        m = myokit.Model()
        c = m.add_component('comp')
        t = c.add_variable('time', rhs=myokit.Number(0))
        t.set_unit(myokit.units.second)
        t.set_binding('time')
        c2 = m.add_component('comp2')
        v = c.add_variable('var', initial_value=3)
        v.set_rhs(myokit.Name(t))
        v2 = c2.add_variable('var', initial_value=3)
        v2.set_rhs(myokit.Name(v))

        # check that the equation is exported correctly to sbml.Model
        s = sbml.Model.from_myokit_model(m)
        parameter_names = [v.sid() for v in s.parameters()]
        self.assertCountEqual(parameter_names, ['comp_var', 'comp2_var'])
        v2_sbml = s.parameter('comp2_var')
        self.assertEqual(
            v2_sbml.value(),
            myokit.Name(v)
        )

        # check that the equation is exported correctly to sbml string
        sbml_str = write_string(s).decode("utf8")
        self.assertIn('<rateRule variable="comp2_var">', sbml_str)
        self.assertIn('<ci>comp_var</ci>', sbml_str)
        self.assertIn('<rateRule variable="comp_var">', sbml_str)
        self.assertIn(
            '<ci>http://www.sbml.org/sbml/symbols/time</ci>',
            sbml_str
        )

    def test_variable_dependent_variables(self):
        m = myokit.Model()
        c = m.add_component('comp')
        t = c.add_variable('time', rhs=myokit.Number(0))
        t.set_unit(myokit.units.second)
        t.set_binding('time')
        v = c.add_variable('var', initial_value=3)
        v.set_rhs(myokit.Number(1))
        v2 = c.add_variable('var2')
        v2.set_rhs(myokit.Name(v))

        s = sbml.Model.from_myokit_model(m)
        self.assertEqual(s.parameter('var2').is_rate(), False)
        self.assertEqual(s.parameter('var').is_rate(), True)

        sbml_str = write_string(s).decode("utf8")
        self.assertIn('<initialAssignment symbol="var">', sbml_str)
        self.assertNotIn('<initialAssignment symbol="var2">', sbml_str)
        self.assertIn('<assignmentRule variable="var2">', sbml_str)



#
# Latex expression writer
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import myokit.formats


class LatexExpressionWriter(myokit.formats.ExpressionWriter):
    """
    This :class:`ExpressionWriter <myokit.formats.ExpressionWriter>` translates
    Myokit :class:`expressions <myokit.Expression>` to their Tex equivalent.
    """
    def __init__(self):
        super().__init__()

        # Default time variable
        self._time = None
        self.set_time_variable_name()

        # Default lhs function
        def flhs(lhs):
            v = r'\text{' + self._prepare_name(lhs.var().uname()) + '}'
            if isinstance(lhs, myokit.Name):
                return v
            elif isinstance(lhs, myokit.Derivative):
                return r'\frac{d' + v + r'}{d\text{' + self._time + '}}'
            elif isinstance(lhs, myokit.PartialDerivative):
                v2 = self._prepare_name(
                    lhs.independent_expression().var().uname())
                return r'\frac{\partial' + v + r'}{\partial\text{' + v2 + '}}'
            elif isinstance(lhs, myokit.InitialValue):
                return v + r'(\text{' + self._time + '} = 0)'
            else:   # pragma: no cover
                raise ValueError(f'Unsupported LHS type {type(lhs)}')

        self._flhs = None
        self.set_lhs_function(flhs)

    def set_lhs_function(self, f):
        """
        Sets a naming function, will be called to get the variable name from a
         :class:`myokit.LhsExpression`. This function is also responsible for
         converting it to a suitable Latex form (e.g. adding `\\text{}`).

        The argument ``f`` should be a function that takes an ``LhsExpression``
        as input and returns a string.
        """
        self._flhs = f

    def set_time_variable_name(self, name='t'):
        """ Sets a name to use for the time variable in derivatives. """
        self._time = self._prepare_name(name)

    def eq(self, eq):
        """ See :meth:`myokit.formats.ExpressionWriter.eq()`. """
        return self.ex(eq.lhs) + ' = ' + self.ex(eq.rhs)

    def ex(self, e):
        """ See :meth:`myokit.formats.ExpressionWriter.ex()`. """
        b = []
        self._ex(e, b)
        return ''.join(b)

    def _prepare_name(self, text):
        """ Sanitises a name for use in latex. """
        return str(text).replace('_', r'\_')

    def _ex(self, e, b):
        try:
            action = self._op_map[type(e)]
        except KeyError:
            raise ValueError('Unknown expression type: ' + str(type(e)))
        action(e, b)

    def _ex_prefix(self, e, b, op):
        """ Handles _ex() for prefix operators. """
        b.append(op)
        if e.bracket(e[0]):
            b.append(r'\left(')
        self._ex(e[0], b)
        if e.bracket(e[0]):
            b.append(r'\right)')

    def _ex_infix(self, e, b, op):
        """ Handles _ex() for infix operators. """
        if e.bracket(e[0]):
            b.append(r'\left(')
        self._ex(e[0], b)
        if e.bracket(e[0]):
            b.append(r'\right)')
        b.append(op)
        if e.bracket(e[1]):
            b.append(r'\left(')
        self._ex(e[1], b)
        if e.bracket(e[1]):
            b.append(r'\right)')

    def _ex_function(self, e, b, func):
        """ Handles _ex() for function operators. """
        b.append(func)
        b.append(r'\left(')
        b.append(','.join([self.ex(x) for x in e]))
        b.append(r'\right)')

    def _ex_infix_condition(self, e, b, op):
        """ Handles _ex() for infix condition operators. """
        b.append(r'\left(')
        self._ex(e[0], b)
        b.append(op)
        self._ex(e[1], b)
        b.append(r'\right)')

    def _ex_name(self, e, b):
        b.append(self._flhs(e))

    def _ex_derivative(self, e, b):
        b.append(self._flhs(e))

    def _ex_partial_derivative(self, e, b):
        b.append(self._flhs(e))

    def _ex_initial_value(self, e, b):
        b.append(self._flhs(e))

    def _ex_number(self, e, b):
        b.append(myokit.float.str(e).strip())
        u = e.unit()
        if u is not None and u is not myokit.units.dimensionless:
            u = str(u)[1:-1]
            b.append(r' \text{' + u + '}')

    def _ex_prefix_plus(self, e, b):
        self._ex_prefix(e, b, '+')

    def _ex_prefix_minus(self, e, b):
        self._ex_prefix(e, b, '-')

    def _ex_plus(self, e, b):
        self._ex_infix(e, b, '+')

    def _ex_minus(self, e, b):
        self._ex_infix(e, b, '-')

    def _ex_multiply(self, e, b):
        self._ex_infix(e, b, r'\cdot')

    def _ex_divide(self, e, b):
        b.append(r'\frac{')
        self._ex(e[0], b)
        b.append('}{')
        self._ex(e[1], b)
        b.append('}')

    def _ex_quotient(self, e, b):
        # Note: Quotient in myokit uses round-to-zero (like Python does!)
        # See: myokit.Quotient
        b.append(r'\left\lfloor')
        self._ex_divide(e, b)
        b.append(r'\right\rfloor')

    def _ex_remainder(self, e, b):
        self._ex_infix(e, b, r'\bmod')

    def _ex_power(self, e, b):
        if e.bracket(e[0]) or isinstance(e[0], myokit.Power):
            b.append(r'\left(')
        self._ex(e[0], b)
        if e.bracket(e[0]) or isinstance(e[0], myokit.Power):
            b.append(r'\right)')
        b.append('^')
        if e.bracket(e[1]) or isinstance(e[1], myokit.Power):
            b.append('{')
        self._ex(e[1], b)
        if e.bracket(e[1]) or isinstance(e[1], myokit.Power):
            b.append('}')

    def _ex_sqrt(self, e, b):
        b.append(r'\sqrt{')
        self._ex(e[0], b)
        b.append('}')

    def _ex_sin(self, e, b):
        self._ex_function(e, b, r'\sin')

    def _ex_cos(self, e, b):
        self._ex_function(e, b, r'\cos')

    def _ex_tan(self, e, b):
        self._ex_function(e, b, r'\tan')

    def _ex_asin(self, e, b):
        self._ex_function(e, b, r'\arcsin')

    def _ex_acos(self, e, b):
        self._ex_function(e, b, r'\arccos')

    def _ex_atan(self, e, b):
        self._ex_function(e, b, r'\arctan')

    def _ex_exp(self, e, b):
        self._ex_function(e, b, r'\exp')

    def _ex_log(self, e, b):
        b.append(r'\log')
        if len(e) > 1:
            b.append('_{')
            self._ex(e[1], b)
            b.append('}')
        b.append(r'\left(')
        self._ex(e[0], b)
        b.append(r'\right)')

    def _ex_log10(self, e, b):
        b.append(r'\log_{10}')
        b.append(r'\left(')
        self._ex(e[0], b)
        b.append(r'\right)')

    def _ex_floor(self, e, b):
        b.append(r'\left\lfloor{')
        self._ex(e[0], b)
        b.append(r'}\right\rfloor')

    def _ex_ceil(self, e, b):
        b.append(r'\left\lceil{')
        self._ex(e[0], b)
        b.append(r'}\right\rceil')

    def _ex_abs(self, e, b):
        b.append(r'\lvert{')
        self._ex(e[0], b)
        b.append(r'}\rvert')

    def _ex_equal(self, e, b):
        self._ex_infix_condition(e, b, '=')

    def _ex_not_equal(self, e, b):
        self._ex_infix_condition(e, b, r'\neq')

    def _ex_more(self, e, b):
        self._ex_infix_condition(e, b, '>')

    def _ex_less(self, e, b):
        self._ex_infix_condition(e, b, '<')

    def _ex_more_equal(self, e, b):
        self._ex_infix_condition(e, b, r'\geq')

    def _ex_less_equal(self, e, b):
        self._ex_infix_condition(e, b, r'\leq')

    def _ex_not(self, e, b):
        b.append(r'\left(\not')
        self._ex(e[0], b)
        b.append(r'\right)')

    def _ex_and(self, e, b):
        self._ex_infix_condition(e, b, r'\and')

    def _ex_or(self, e, b):
        self._ex_infix_condition(e, b, r'\or')

    def _ex_if(self, e, b):
        # Not suported
        self._ex_function(e, b, r'\text{if}')

    def _ex_piecewise(self, e, b):
        # Not suported
        self._ex_function(e, b, r'\text{piecewise}')



#!/usr/bin/env python3
#
# Tests writing of AnsiC expressions, as used by the Simulation.
# C++ tests are lumped in as well.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import math
import unittest

import myokit
import myokit.formats.ansic
import myokit.formats.cpp

from myokit import (
    Number, PrefixPlus, PrefixMinus, Plus, Minus,
    Multiply, Divide, Quotient, Remainder, Power, Sqrt,
    Exp, Log, Log10, Sin, Cos, Tan, ASin, ACos, ATan, Floor, Ceil, Abs,
    Not, And, Or, Equal, NotEqual, More, Less, MoreEqual, LessEqual,
    If, Piecewise,
)

import myokit.tests


class AnsiCExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """
    Test conversion to Ansi C, as used by the Simulation.
    Numerical tests are provided by composing and evaluating a single RHS.
    """
    _name = 'ansic'
    _target = myokit.formats.ansic.AnsiCExpressionWriter

    def test_number(self):
        self.eq(Number(1), '1.0')
        self.eq(Number(-2), '-2.0')
        self.eq(Number(13, 'mV'), '13.0')

    def test_name(self):
        self.eq(self.a, 'a')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        self.eq(myokit.PartialDerivative(self.a, self.b), 'partial(a, b)')

    def test_initial_value(self):
        self.eq(myokit.InitialValue(self.a), 'initial(a)')

    def test_prefix_plus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixPlus(p), '+11.0')
        self.eq(PrefixPlus(PrefixPlus(p)), '+(+11.0)')
        self.eq(PrefixPlus(PrefixPlus(PrefixPlus(p))), '+(+(+11.0))')
        self.eq(PrefixPlus(Number('+1')), '+1.0')

        # Test with operators of precedence SUM, PRODUCT
        a, b, c = self.abc
        self.eq(PrefixPlus(Plus(a, b)), '+(a + b)')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(PrefixPlus(Divide(b, a)), '+(b / a)')

    def test_prefix_minus(self):
        # Test with numbers
        p = Number(11, 'kV')
        self.eq(PrefixMinus(p), '-11.0')
        self.eq(PrefixMinus(PrefixMinus(p)), '-(-11.0)')
        self.eq(PrefixMinus(Number(-1)), '-(-1.0)')
        self.eq(PrefixMinus(PrefixMinus(Number(-2))), '-(-(-2.0))')

        # Test with operators of precedence SUM, PRODUCT
        a, b, c = self.abc
        self.eq(PrefixMinus(Minus(a, b)), '-(a - b)')
        self.eq(Multiply(PrefixMinus(Plus(b, a)), c), '-(b + a) * c')
        self.eq(PrefixMinus(Divide(b, a)), '-(b / a)')

    def test_plus_minus(self):
        a, b, c = self.abc
        self.eq(Plus(a, b), 'a + b')
        self.eq(Plus(Plus(a, b), c), 'a + b + c')
        self.eq(Plus(a, Plus(b, c)), 'a + (b + c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Minus(Minus(a, b), c), 'a - b - c')
        self.eq(Minus(a, Minus(b, c)), 'a - (b - c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Plus(Minus(a, b), c), 'a - b + c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')

    def test_multiply_divide(self):
        a, b, c = self.abc
        self.eq(Multiply(a, b), 'a * b')
        self.eq(Multiply(Multiply(a, b), c), 'a * b * c')
        self.eq(Multiply(a, Multiply(b, c)), 'a * (b * c)')

        self.eq(Divide(a, b), 'a / b')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')

        self.eq(Divide(Multiply(a, b), c), 'a * b / c')
        self.eq(Multiply(Divide(a, b), c), 'a / b * c')
        self.eq(Divide(a, Multiply(b, c)), 'a / (b * c)')
        self.eq(Multiply(a, Divide(b, c)), 'a * (b / c)')

        self.eq(Multiply(Minus(a, b), c), '(a - b) * c')
        self.eq(Multiply(a, Plus(b, c)), 'a * (b + c)')
        self.eq(Minus(Multiply(a, b), c), 'a * b - c')
        self.eq(Plus(a, Multiply(b, c)), 'a + b * c')
        self.eq(Divide(Plus(a, b), c), '(a + b) / c')
        self.eq(Divide(a, Minus(b, c)), 'a / (b - c)')
        self.eq(Plus(Divide(a, b), c), 'a / b + c')
        self.eq(Minus(a, Divide(b, c)), 'a - b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')

    def test_quotient(self):
        a, b, c = self.abc
        self.eq(Quotient(a, b), 'floor(a / b)')
        self.eq(Quotient(Plus(a, c), b), 'floor((a + c) / b)')
        self.eq(Quotient(Divide(a, c), b), 'floor(a / c / b)')
        self.eq(Quotient(a, Divide(b, c)), 'floor(a / (b / c))')
        self.eq(Multiply(Quotient(a, b), c), 'floor(a / b) * c')
        # Bracket() method expects a PRODUCT level operation, so will add
        # unnecessary brackets here
        self.eq(Multiply(c, Quotient(a, b)), 'c * (floor(a / b))')

    def test_remainder(self):
        a, b, c = self.abc
        self.eq(Remainder(a, b), '(a - b * floor(a / b))')
        self.eq(Remainder(Plus(a, c), b), '(a + c - b * floor((a + c) / b))')
        self.eq(Multiply(Remainder(a, b), c), '(a - b * floor(a / b)) * c')
        # Bracket() method expects a PRODUCT level operation, so will add
        # unnecessary brackets here
        self.eq(Divide(c, Remainder(b, a)), 'c / ((b - a * floor(b / a)))')

    def test_power(self):
        a, b, c = self.abc
        self.eq(Power(a, b), 'pow(a, b)')
        self.eq(Power(Power(a, b), c), 'pow(pow(a, b), c)')
        self.eq(Power(a, Power(b, c)), 'pow(a, pow(b, c))')

        self.eq(Power(Plus(a, b), c), 'pow(a + b, c)')
        self.eq(Power(a, Minus(b, c)), 'pow(a, b - c)')
        self.eq(Power(Multiply(a, b), c), 'pow(a * b, c)')
        self.eq(Power(a, Divide(b, c)), 'pow(a, b / c)')

    def test_log(self):
        a, b = self.ab
        self.eq(Log(a), 'log(a)')
        self.eq(Log10(a), 'log10(a)')
        self.eq(Log(a, b), '(log(a) / log(b))')

    def test_functions(self):
        a, b = self.ab

        self.eq(Sqrt(a), 'sqrt(a)')
        self.eq(Exp(a), 'exp(a)')
        self.eq(Sin(a), 'sin(a)')
        self.eq(Cos(a), 'cos(a)')
        self.eq(Tan(a), 'tan(a)')
        self.eq(ASin(a), 'asin(a)')
        self.eq(ACos(a), 'acos(a)')
        self.eq(ATan(a), 'atan(a)')
        self.eq(Floor(a), 'floor(a)')
        self.eq(Ceil(a), 'ceil(a)')
        self.eq(Abs(a), 'fabs(a)')

    def test_conditions(self):
        a, b, c, d = self.abcd
        p = myokit.Equal(a, b)
        q = myokit.NotEqual(c, d)

        self.eq(And(p, q), '((a == b) && (c != d))')
        self.eq(Or(q, p), '((c != d) || (a == b))')
        self.eq(Not(And(p, q)), '(!((a == b) && (c != d)))')
        self.eq(Not(p), '(!(a == b))')

        self.eq(Equal(a, b), '(a == b)')
        self.eq(NotEqual(a, b), '(a != b)')
        self.eq(More(b, a), '(b > a)')
        self.eq(Less(d, c), '(d < c)')
        self.eq(MoreEqual(c, a), '(c >= a)')
        self.eq(LessEqual(b, d), '(b <= d)')

        self.eq(And(Equal(a, b), NotEqual(c, d)), '((a == b) && (c != d))')
        self.eq(Or(More(d, c), Less(b, a)), '((d > c) || (b < a))')

    def test_conditionals(self):

        a, b, c, d = self.abcd
        self.eq(If(Equal(a, b), d, c), '((a == b) ? d : c)')
        self.eq(Piecewise(NotEqual(d, c), b, a), '((d != c) ? b : a)')
        self.eq(Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
                '((a == b) ? c : ((a == d) ? 3.0 : 4.0))')

        # Using if-then-else function
        w = self._target()
        w.set_lhs_function(lambda v: v.var().name())
        w.set_condition_function('ite')

        self.assertEqual(w.ex(If(More(b, a), c, d)), 'ite((b > a), c, d)')
        self.assertEqual(w.ex(Piecewise(Less(d, c), b, a)),
                         'ite((d < c), b, a)')
        self.assertEqual(w.ex(
            Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4))),
            'ite((a == b), c, ite((a == d), 3.0, 4.0))')

    def test_in_c(self):
        """ Compile and test the values evaluated in C. """

        class CTester():
            def __init__(self, parent):
                self._parent = parent
                self._m = myokit.Model()
                self._c = self._m.add_component('c')
                t = self._c.add_variable('time', rhs=0, binding='time')
                self._expected = []

            def add(self, expression, expected):
                v = self._c.add_variable(
                    f'v{len(self._expected)}', rhs=expression, initial_value=0)
                self._expected.append(float(expected))

            def run(self):
                s = myokit.Simulation(self._m)
                x = s.evaluate_derivatives()
                self._parent.assertEqual(len(x), len(self._expected))
                for a, b in zip(x, self._expected):
                    self._parent.assertEqual(a, b)

        c = CTester(self,)

        c.add(Number(12, 'pF'), 12)

        c.add(PrefixPlus(Number(3)), 3)
        c.add(PrefixPlus(PrefixPlus(PrefixPlus(Number(4)))), 4)
        c.add(PrefixMinus(Number(6)), -6)
        c.add(PrefixMinus(PrefixMinus(Number(2))), 2)
        c.add(PrefixMinus(PrefixMinus(PrefixMinus(Number(5)))), -5)
        c.add(PrefixMinus(PrefixMinus(PrefixMinus(Number(-5)))), 5)

        c.add(Plus(Number(4), Number(2)), 6)
        c.add(Minus(Number(5), Number(1.5)), 3.5)

        c.add(Multiply(Number(7), Number(9)), 63)
        c.add(Divide(Number(5), Number(2)), 2.5)
        c.add(Divide(Divide(Number(12), Number(2)), Number(2)), 3)
        c.add(Divide(Number(1), Divide(Number(2), Number(3))), 1.5)

        c.add(Remainder(Number(10), Number(4)), 2)
        c.add(Remainder(Number(10), Number(6)), 4)
        c.add(Remainder(Number(5), Number(3)), 2)
        c.add(Remainder(Number(-5), Number(3)), 1)
        c.add(Remainder(Number(5), Number(-3)), -1)
        c.add(Remainder(Number(-5), Number(-3)), -2)

        c.add(Quotient(Number(10), Number(4)), 2)
        c.add(Quotient(Number(10), Number(6)), 1)
        c.add(Quotient(Number(5), Number(3)), 1)
        c.add(Quotient(Number(-5), Number(3)), -2)
        c.add(Quotient(Number(5), Number(-3)), -2)
        c.add(Quotient(Number(-5), Number(-3)), 1)

        c.add(Power(Number(4), Power(Number(2), Number(3))), 65536)
        c.add(Power(Power(Number(2), Number(3)), Number(4)), 4096)

        c.add(Log(Number(3)), math.log(3))
        c.add(Log10(Number(1000)), 3)
        c.add(Log10(Number(0.01)), -2)
        c.add(Log(Number(27), Number(3)), 3)
        c.add(Divide(Number(12), Log(Number(256), Number(4))), 3)
        c.add(Divide(Log(Number(256), Number(4)), Number(4)), 1)

        c.add(Sqrt(Number(9)), 3)
        c.add(Exp(Number(3)), math.exp(3))
        c.add(Sin(Number(1)), math.sin(1))
        c.add(Cos(Number(1)), math.cos(1))
        c.add(Tan(Number(4)), math.tan(4))
        c.add(ASin(Number(0.4)), math.asin(0.4))
        c.add(ACos(Number(0.4)), math.acos(0.4))
        c.add(ATan(Number(0.4)), math.atan(0.4))

        c.add(Floor(Number(3.9)), 3)
        c.add(Floor(Number(-3.9)), -4)
        c.add(Ceil(Number(4.01)), 5)
        c.add(Ceil(Number(-4.01)), -4)
        c.add(Abs(PrefixMinus(Number(12))), 12)
        c.add(Abs(Number(-13)), 13)

        true = Equal(Number(2), Number(2))
        false = Equal(Number(3), Number(-1))
        a, b = Number(10), Number(20)
        c.add(If(true, a, b), 10)
        c.add(If(false, a, b), 20)

        c.add(If(More(Number(5), Number(3)), a, b), 10)
        c.add(If(More(Number(3), Number(3)), a, b), 20)
        c.add(If(MoreEqual(Number(3), Number(3)), a, b), 10)
        c.add(If(Less(Number(3), Number(5)), a, b), 10)
        c.add(If(More(Number(3), Number(3)), a, b), 20)
        c.add(If(LessEqual(Number(3), Number(3)), a, b), 10)

        c.add(If(And(false, false), a, b), 20)
        c.add(If(And(false, true), a, b), 20)
        c.add(If(And(true, false), a, b), 20)
        c.add(If(And(true, true), a, b), 10)
        c.add(If(Or(false, false), a, b), 20)
        c.add(If(Or(false, true), a, b), 10)
        c.add(If(Or(true, false), a, b), 10)
        c.add(If(Or(true, true), a, b), 10)
        c.add(If(Not(true), a, b), 20)
        c.add(If(Not(false), a, b), 10)

        c.add(Piecewise(true, Number(10), false, Number(20), Number(30)), 10)
        c.add(Piecewise(true, Number(10), true, Number(20), Number(30)), 10)
        c.add(Piecewise(false, Number(10), true, Number(20), Number(30)), 20)
        c.add(Piecewise(false, Number(10), false, Number(20), Number(30)), 30)

        c.run()


if __name__ == '__main__':
    unittest.main()



#
# SBML Writer: Writes an SBML Model to disk
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
from typing import Tuple
from lxml import etree

import myokit
from myokit._unit import Quantity
from myokit.formats.mathml._ewriter import MathMLExpressionWriter
from myokit.formats.sbml._api import (
    Model, Compartment,
    Parameter, Species,
    Reaction, SpeciesReference
)


def write_file(path: str, model: Model):
    """
    Writes an SBML model to the given path.
    """
    return SBMLWriter.write_file(path, model)


def write_string(model: Model) -> str:
    """
    Writes an SBML model to a string and returns it.
    """
    return SBMLWriter.write_string(model)


class SBMLWriter:
    """
    Writes SBML documents
    """

    @staticmethod
    def write_file(path: str, model: Model):
        tree = SBMLWriter._model(model)
        # Write to disk
        tree.write(
            path,
            encoding='utf-8',
            method='xml',
            xml_declaration=True,
            pretty_print=True,
        )

    @staticmethod
    def write_string(model: Model) -> str:
        tree = SBMLWriter._model(model)
        return etree.tostring(
            tree,
            encoding='utf-8',
            method='xml',
            pretty_print=True,
        )

    @staticmethod
    def _compartment(
            compartment: Compartment,
            unit_to_str: dict
    ) -> etree.Element:
        node = etree.Element('compartment', id=compartment.sid())
        if compartment.size_units() != myokit.units.dimensionless:
            node.attrib['units'] = unit_to_str[compartment.size_units()]
        if compartment.spatial_dimensions() is not None:
            node.attrib['spatialDimensions'] = str(
                compartment.spatial_dimensions()
            )
        return node

    @staticmethod
    def _unit(sid: str, unit: myokit.Unit) -> etree.Element:
        node = etree.Element('unitDefinition', id=sid)
        kinds = [
            'gram', 'metre', 'second',
            'ampere', 'kelvin', 'candela', 'mole'
        ]
        multiplier = unit.multiplier()
        for exponent, kind in zip(unit.exponents(), kinds):
            if exponent != 0:
                child = etree.Element('unit')
                child.attrib['kind'] = kind
                child.attrib['exponent'] = str(exponent)
                if multiplier is not None:
                    child.attrib['multiplier'] = str(multiplier)
                    multiplier = None
                node.append(child)
        # might also have a dimensionless unit and a multiplier
        if multiplier is not None:
            child = etree.Element('unit')
            child.attrib['kind'] = 'dimensionless'
            child.attrib['multiplier'] = str(multiplier)
            node.append(child)
        return node

    @staticmethod
    def _parameter(
            parameter: Parameter,
            unit_to_str_map: dict
    ) -> Tuple[etree.Element, etree.Element, etree.Element]:
        """
        returns the XML representation of this parameter as a tuple of
        (parameter, initial_assignment, rule).
        """
        parameter_xml = etree.Element('parameter', id=parameter.sid())
        if (
            parameter.units() is not None and
            parameter.units() != myokit.units.dimensionless
        ):
            parameter_xml.attrib['units'] = unit_to_str_map[parameter.units()]

        if parameter.is_constant():
            parameter_xml.attrib['constant'] = 'true'

        if parameter.is_literal():
            value = parameter.value().eval()
            parameter_xml.attrib['value'] = str(value)
            return parameter_xml, None, None
        else:
            initial_assignment, rule = SBMLWriter._quantity(parameter)
            return parameter_xml, initial_assignment, rule

    @staticmethod
    def _math(expression: myokit.Expression) -> etree.Element:
        math = etree.Element(
            'math',
            xmlns='http://www.w3.org/1998/Math/MathML'
        )

        def flhs(lhs):
            var = lhs.var()
            if isinstance(var, str):
                return var
            if var.binding() == 'time':
                return "http://www.sbml.org/sbml/symbols/time"
            return var.uname()

        mathml_writer = MathMLExpressionWriter()
        mathml_writer.set_lhs_function(flhs)
        mathml_writer.ex(expression, math)
        return math

    @staticmethod
    def _quantity(quantity: Quantity) -> etree.Element:
        initial_value = quantity.initial_value()
        initial_assignment = None
        if initial_value is not None:
            initial_assignment = etree.Element(
                'initialAssignment', symbol=quantity.sid()
            )
            math = SBMLWriter._math(initial_value)
            initial_assignment.append(math)

        value = quantity.value()
        rule = None
        if value is not None:
            if quantity.is_rate():
                rule_type = 'rateRule'
            else:
                rule_type = 'assignmentRule'
            rule = etree.Element(rule_type, variable=quantity.sid())
            math = SBMLWriter._math(quantity.value())
            rule.append(math)
        return initial_assignment, rule

    @staticmethod
    def _reaction(reaction: Reaction) -> etree.Element:
        reaction_xml = etree.Element('reaction', id=reaction.sid())
        list_of_reactants = etree.Element('listOfReactants')
        for reactant in reaction.reactants():
            node = SBMLWriter._species_reference(reactant)
            list_of_reactants.append(node)
        reaction_xml.append(list_of_reactants)
        list_of_products = etree.Element('listOfProducts')
        for product in reaction.products():
            node = SBMLWriter._species_reference(product)
            list_of_products.append(node)
        reaction_xml.append(list_of_products)
        list_of_modifiers = etree.Element('listOfModifiers')
        for modifier in reaction.modifiers():
            node = SBMLWriter._modifier_species_reference(modifier)
            list_of_modifiers.append(node)
        reaction_xml.append(list_of_modifiers)
        if reaction.kinetic_law() is not None:
            kinetic_law = etree.Element('kineticLaw')
            math = SBMLWriter._math(reaction.kinetic_law())
            kinetic_law.append(math)
            reaction_xml.append(kinetic_law)
        return reaction_xml

    @staticmethod
    def _species(species: Species, unit_to_str_map: dict) -> Tuple[
        etree.Element, etree.Element
    ]:
        """
        Returns the XML representation of this species as a tuple of
        (species, rule).
        """
        species_xml = etree.Element('species', id=species.sid())
        species_xml.attrib['compartment'] = species.compartment().sid()
        initial_value, initial_value_in_amount = species.initial_value()
        if initial_value_in_amount is None:
            if species.is_amount():
                attrib_name = 'initialAmount'
            else:
                attrib_name = 'initialConcentration'
        else:
            if initial_value_in_amount:
                attrib_name = 'initialAmount'
            else:
                attrib_name = 'initialConcentration'
        if initial_value is not None:
            initial_value_eval = initial_value.eval()
            species_xml.attrib[attrib_name] = str(initial_value_eval)
        species_xml.attrib['constant'] = str(species.is_constant())
        if species.substance_units() != myokit.units.dimensionless:
            species_xml.attrib['units'] = unit_to_str_map[
                species.substance_units()
            ]
        species_xml.attrib['boundaryCondition'] = str(species.is_boundary())

        if species.value() is None:
            return species_xml, None

        if species.is_rate():
            rule_type = 'rateRule'
        else:
            rule_type = 'assignmentRule'
        rule = etree.Element(rule_type, variable=species.sid())
        math = SBMLWriter._math(species.value())
        rule.append(math)
        return species_xml, rule

    @staticmethod
    def _species_reference(ref: SpeciesReference) -> etree.Element:
        species_reference = etree.Element(
            'speciesReference',
            species=ref.species().sid()
        )
        if ref.sid() is not None:
            species_reference.attrib['id'] = ref.sid()
        value = ref.value()
        if value is not None:
            value_eval = value.eval()
            species_reference.attrib['stoichiometry'] = str(value_eval)
        return species_reference

    @staticmethod
    def _modifier_species_reference(ref: SpeciesReference) -> etree.Element:
        species_reference = etree.Element(
            'modifierSpeciesReference',
            species=ref.species().sid()
        )
        if ref.sid() is not None:
            species_reference.attrib['id'] = ref.sid()
        return species_reference

    @staticmethod
    def _model(model: Model) -> etree.ElementTree:
        root = etree.Element(
            'sbml',
            xmlns='http://www.sbml.org/sbml/level3/version2/core',
            level='3',
            version='2'
        )

        # setup a map from unit to string
        unit_map_to_str = {
            unit: string for string, unit in Model.base_units.items()
        }
        for unitid, unit in model.units().items():
            unit_map_to_str[unit] = unitid

        name = model.name() if model.name() else 'unnamed_model'
        model_root = etree.Element('model', id=name)
        if model.time_units() != myokit.units.dimensionless:
            model_root.attrib['timeUnits'] = unit_map_to_str[
                model.time_units()
            ]
        if model.area_units() != myokit.units.dimensionless:
            model_root.attrib['areaUnits'] = unit_map_to_str[
                model.area_units()
            ]
        if model.length_units() != myokit.units.dimensionless:
            model_root.attrib['lengthUnits'] = unit_map_to_str[
                model.length_units()
            ]
        if model.substance_units() != myokit.units.dimensionless:
            model_root.attrib['substanceUnits'] = unit_map_to_str[
                model.substance_units()
            ]
        if model.extent_units() != myokit.units.dimensionless:
            model_root.attrib['extentUnits'] = unit_map_to_str[
                model.extent_units()
            ]
        if model.volume_units() != myokit.units.dimensionless:
            model_root.attrib['volumeUnits'] = unit_map_to_str[
                model.volume_units()
            ]

        if model.has_units():
            list_of_units = etree.Element('listOfUnitDefinitions')
            for sid, unit in model.units().items():
                node = SBMLWriter._unit(sid, unit)
                list_of_units.append(node)
            model_root.append(list_of_units)
        if model.compartments():
            list_of_compartments = etree.Element('listOfCompartments')
            for compartment in model.compartments():
                node = SBMLWriter._compartment(compartment, unit_map_to_str)
                list_of_compartments.append(node)
            model_root.append(list_of_compartments)
        list_of_rules = etree.Element('listOfRules')
        list_of_initial_assignments = etree.Element('listOfInitialAssignments')
        if model.parameters():
            list_of_parameters = etree.Element('listOfParameters')
            for parameter in model.parameters():
                param_node, initial_value_node, rule_node = \
                    SBMLWriter._parameter(parameter, unit_map_to_str)
                list_of_parameters.append(param_node)
                if initial_value_node is not None:
                    list_of_initial_assignments.append(initial_value_node)
                if rule_node is not None:
                    list_of_rules.append(rule_node)
            model_root.append(list_of_parameters)
        if model.species_list():
            list_of_species = etree.Element('listOfSpecies')
            for species in model.species_list():
                species_node, rule_node = SBMLWriter._species(
                    species,
                    unit_map_to_str
                )
                if rule_node is not None:
                    list_of_rules.append(rule_node)
                list_of_species.append(species_node)
            model_root.append(list_of_species)
        if model.reactions():
            list_of_reactions = etree.Element('listOfReactions')
            for reaction in model.reactions():
                node = SBMLWriter._reaction(reaction)
                list_of_reactions.append(node)
            model_root.append(list_of_reactions)

        if len(list_of_initial_assignments) > 0:
            model_root.append(list_of_initial_assignments)

        if len(list_of_rules) > 0:
            model_root.append(list_of_rules)

        root.append(model_root)
        return etree.ElementTree(root)



#!/usr/bin/env python3
#
# Tests the EasyML module.
#
# This file is part of Myokit.
# See http://myokit.org for copyright, sharing, and licensing details.
#
import os
import unittest

import myokit
import myokit.formats
import myokit.formats.easyml
import myokit.tests

from myokit import (
    Number, PrefixPlus, PrefixMinus, Plus, Minus,
    Multiply, Divide, Quotient, Remainder, Power, Sqrt,
    Exp, Log, Log10, Sin, Cos, Tan, ASin, ACos, ATan, Floor, Ceil, Abs,
    Not, And, Or, Equal, NotEqual, More, Less, MoreEqual, LessEqual,
    If, Piecewise,
)

from myokit.tests import TemporaryDirectory, WarningCollector, DIR_DATA


# Model that requires unit conversion
units_model = """
[[model]]
membrane.V = -0.08
hh.x = 0.1
hh.y = 0.9
mm.C = 0.9

[engine]
time = 0 [s]
    in [s]
    bind time

[membrane]
dot(V) = (hh.I1 + mm.I2) / C
    in [V]
C = 20 [pF]
    in [pF]

[hh]
dot(x) = (inf - x) / tau
    inf = 0.8
    tau = 3 [s]
        in [s]
dot(y) = alpha * (1 - y) - beta * y
    alpha = 0.1 [1/s]
        in [1/s]
    beta = 0.2 [1/s]
        in [1/s]
I1 = 3 [pS] * x * y * (membrane.V - 0.05 [V])
    in [pA]

[mm]
dot(C) = beta * O - alpha * C
alpha = 0.3 [1/s]
    in [1/s]
beta = 0.4 [1/s]
    in [1/s]
O = 1 - C
I2 = 2 [pS] * O * (membrane.V + 0.02 [V])
    in [pA]
"""
units_output = """
/*
This file was generated by Myokit.
*/

V; .nodal(); .external(Vm);
Iion; .nodal(); .external();

V_init = -80.0;
x_init = 0.1;
y_init = 0.9;
C_init = 0.9;

// hh
I1 = 3.0 * x * y * (V / 1000.0 - 0.05) * 0.05;
x_inf = 0.8;
tau_x = 3.0 * 1000.0;
alpha_y = 0.1 * 0.001;
beta_y = 0.2 * 0.001;

// mm
diff_C = (mm_beta * O - mm_alpha * C) * 0.001;
I2 = 2.0 * O * (V / 1000.0 + 0.02) * 0.05;
O = 1.0 - C;
mm_alpha = 0.3;
mm_beta = 0.4;

// Sum of currents
Iion = I1 + I2;

// Markov model
group {
  C;
}.method(markov_be);

// Trace all currents and state variables
group {
  I1;
  I2;
  V;
  x;
  y;
  C;
}.trace();

// Parameters
group {
  mm_alpha;
  mm_beta;
}.param();
"""


class EasyMLExporterTest(unittest.TestCase):
    """ Tests EasyML export. """

    def test_easyml_exporter(self):
        # Tests exporting a model

        model = myokit.load_model('example')
        with TemporaryDirectory() as d:
            path = d.path('easy.model')

            # Test with simple model
            e = myokit.formats.easyml.EasyMLExporter()
            e.model(path, model)

            # Test with extra bound variables
            model.get('membrane.C').set_binding('hello')
            e.model(path, model)

            # Test without V being a state variable
            v = model.get('membrane.V')
            v.demote()
            v.set_rhs(3)
            e.model(path, model)

            # Test with invalid model
            v.set_rhs('2 * V')
            self.assertRaisesRegex(
                myokit.ExportError, 'valid model', e.model, path, model)

    def test_easyml_exporter_static(self):
        # Tests exporting a model (with HH and markov states) and compares
        # against reference output.

        # Export model
        m = myokit.load_model(os.path.join(DIR_DATA, 'decker-2009.mmt'))
        e = myokit.formats.easyml.EasyMLExporter()
        with TemporaryDirectory() as d:
            path = d.path('decker.model')
            e.model(path, m)
            with open(path, 'r') as f:
                observed = f.readlines()

        # Load expected output
        with open(os.path.join(DIR_DATA, 'decker.model'), 'r') as f:
            expected = f.readlines()

        # Compare (line by line, for readable output)
        for ob, ex in zip(observed, expected):
            self.assertEqual(ob, ex)
        self.assertEqual(len(observed), len(expected))

    def test_unit_conversion(self):
        # Tests exporting a model that requires unit conversion

        # Export model
        m = myokit.parse_model(units_model)
        e = myokit.formats.easyml.EasyMLExporter()
        with TemporaryDirectory() as d:
            path = d.path('easy.model')
            e.model(path, m)
            with open(path, 'r') as f:
                observed = f.read().strip().splitlines()

        # Get expected output
        expected = units_output.strip().splitlines()

        # Compare (line by line, for readable output)
        for ob, ex in zip(observed, expected):
            self.assertEqual(ob, ex)
        self.assertEqual(len(observed), len(expected))

        # Test warnings are raised if conversion fails
        m.get('membrane.V').set_rhs('hh.I1 + mm.I2')
        m.get('membrane').remove_variable(m.get('membrane.C'))
        with TemporaryDirectory() as d:
            path = d.path('easy.model')
            with WarningCollector() as c:
                e.model(path, m)
            self.assertIn('Unable to convert hh.I1', c.text())
            self.assertIn('Unable to convert mm.I2', c.text())

        m.get('engine.time').set_unit(myokit.units.cm)
        with TemporaryDirectory() as d:
            path = d.path('easy.model')
            with WarningCollector() as c:
                e.model(path, m)
            self.assertIn('Unable to convert time units [cm]', c.text())

    def test_export_reused_variable(self):
        # Tests exporting when an `inf` or other special variable is used twice

        # Create model re-using tau and inf
        m = myokit.parse_model(
            """
            [[model]]
            m.V = -80
            c.x = 0.1
            c.y = 0.1

            [m]
            time = 0 bind time
            i_ion = c.I
            dot(V) = -i_ion

            [c]
            inf = 0.5
            tau = 3
            dot(x) = (inf - x) / tau
            dot(y) = (inf - y) / tau
            I = x * y * (m.V - 50)
            """)

        # Export, and read back in
        e = myokit.formats.easyml.EasyMLExporter()
        with TemporaryDirectory() as d:
            path = d.path('easy.model')
            e.model(path, m)
            with open(path, 'r') as f:
                x = f.read()

        self.assertIn('x_inf =', x)
        self.assertIn('y_inf =', x)
        self.assertIn('tau_x =', x)
        self.assertIn('tau_y =', x)

    def test_easyml_exporter_fetching(self):
        # Tests getting an EasyML exporter via the 'exporter' interface

        e = myokit.formats.exporter('easyml')
        self.assertIsInstance(e, myokit.formats.easyml.EasyMLExporter)

    def test_capability_reporting(self):
        # Tests if the correct capabilities are reported
        e = myokit.formats.easyml.EasyMLExporter()
        self.assertTrue(e.supports_model())


class EasyMLExpressionWriterTest(myokit.tests.ExpressionWriterTestCase):
    """ Test conversion to EasyML syntax. """
    _name = 'easyml'
    _target = myokit.formats.easyml.EasyMLExpressionWriter

    def test_number(self):
        self.eq(Number(1), '1.0')
        self.eq(Number(-2), '-2.0')
        self.eq(Number(13, 'mV'), '13.0')

    def test_name(self):
        # Inherited from CBasedExpressionWriter
        self.eq(self.a, 'a')
        w = self._target()
        w.set_lhs_function(lambda v: v.var().qname().upper())
        self.assertEqual(w.ex(self.a), 'COMP.A')

    def test_derivative(self):
        # Inherited from CBasedExpressionWriter
        self.eq(myokit.Derivative(self.a), 'dot(a)')

    def test_partial_derivative(self):
        e = myokit.PartialDerivative(self.a, self.b)
        self.assertRaisesRegex(NotImplementedError, 'Partial', self.w.ex, e)

    def test_initial_value(self):
        e = myokit.InitialValue(self.a)
        self.assertRaisesRegex(NotImplementedError, 'Initial', self.w.ex, e)

    def test_prefix_plus_minus(self):
        # Inherited from CBasedExpressionWriter
        p = Number(11, 'kV')
        a, b, c = self.abc
        self.eq(PrefixPlus(p), '+11.0')
        self.eq(PrefixPlus(PrefixPlus(PrefixPlus(p))), '+(+(+11.0))')
        self.eq(Divide(PrefixPlus(Plus(a, b)), c), '+(a + b) / c')
        self.eq(PrefixMinus(p), '-11.0')
        self.eq(PrefixMinus(PrefixMinus(p)), '-(-11.0)')
        self.eq(PrefixMinus(Number(-1)), '-(-1.0)')
        self.eq(PrefixMinus(Minus(a, b)), '-(a - b)')
        self.eq(Multiply(PrefixMinus(Plus(b, a)), c), '-(b + a) * c')
        self.eq(PrefixMinus(Divide(b, a)), '-(b / a)')

    def test_plus_minus(self):
        a, b, c = self.abc
        self.eq(Plus(a, b), 'a + b')
        self.eq(Plus(Plus(a, b), c), 'a + b + c')
        self.eq(Plus(a, Plus(b, c)), 'a + (b + c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Minus(Minus(a, b), c), 'a - b - c')
        self.eq(Minus(a, Minus(b, c)), 'a - (b - c)')

        self.eq(Minus(a, b), 'a - b')
        self.eq(Plus(Minus(a, b), c), 'a - b + c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')
        self.eq(Minus(Plus(a, b), c), 'a + b - c')
        self.eq(Minus(a, Plus(b, c)), 'a - (b + c)')

        # Substitution of expm for 1 - exp() and exp() - 1
        self.eq(Minus(Exp(Number(2)), Number(1)), 'expm1(2.0)')
        self.eq(Minus(Number(1), Exp(Number(3))), '-expm1(3.0)')

    def test_multiply_divide(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        self.eq(Multiply(a, b), 'a * b')
        self.eq(Multiply(Multiply(a, b), c), 'a * b * c')
        self.eq(Multiply(a, Multiply(b, c)), 'a * (b * c)')
        self.eq(Divide(a, b), 'a / b')
        self.eq(Divide(Divide(a, b), c), 'a / b / c')
        self.eq(Divide(a, Divide(b, c)), 'a / (b / c)')

    def test_quotient(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        with WarningCollector():
            self.eq(Quotient(a, b), 'floor(a / b)')
            self.eq(Quotient(Plus(a, c), b), 'floor((a + c) / b)')
            self.eq(Quotient(Divide(a, c), b), 'floor(a / c / b)')
            self.eq(Quotient(a, Divide(b, c)), 'floor(a / (b / c))')
            self.eq(Multiply(Quotient(a, b), c), 'floor(a / b) * c')
            self.eq(Multiply(c, Quotient(a, b)), 'c * (floor(a / b))')

    def test_remainder(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        with WarningCollector():
            self.eq(Remainder(a, b), '(a - b * floor(a / b))')
            self.eq(Remainder(Plus(a, c), b),
                    '(a + c - b * floor((a + c) / b))')
            self.eq(Multiply(Remainder(a, b), c), '(a - b * floor(a / b)) * c')
            self.eq(Divide(c, Remainder(b, a)), 'c / ((b - a * floor(b / a)))')

    def test_power(self):
        # Inherited from CBasedExpressionWriter
        a, b, c = self.abc
        self.eq(Power(a, b), 'pow(a, b)')
        self.eq(Power(Power(a, b), c), 'pow(pow(a, b), c)')
        self.eq(Power(a, Power(b, c)), 'pow(a, pow(b, c))')

    def test_log(self):
        # Inherited from CBasedExpressionWriter
        a, b = self.ab
        self.eq(Log(a), 'log(a)')
        self.eq(Log10(a), 'log10(a)')
        self.eq(Log(a, b), '(log(a) / log(b))')

    def test_functions(self):
        a, b = self.ab

        self.eq(Sqrt(a), 'sqrt(a)')
        self.eq(Exp(a), 'exp(a)')
        self.eq(Cos(a), 'cos(a)')
        self.eq(ACos(a), 'acos(a)')
        self.eq(Abs(a), 'fabs(a)')

        with WarningCollector() as c:
            self.eq(Sin(a), 'sin(a)')
        self.assertIn('Unsupported', c.text())

        with WarningCollector() as c:
            self.eq(Tan(a), 'tan(a)')
        self.assertIn('Unsupported', c.text())

        with WarningCollector() as c:
            self.eq(ASin(a), 'asin(a)')
        self.assertIn('Unsupported', c.text())

        with WarningCollector() as c:
            self.eq(ATan(a), 'atan(a)')
        self.assertIn('Unsupported', c.text())

        with WarningCollector() as c:
            self.eq(Floor(a), 'floor(a)')
        self.assertIn('Unsupported', c.text())

        with WarningCollector() as c:
            self.eq(Ceil(a), 'ceil(a)')
        self.assertIn('Unsupported', c.text())

    def test_conditions(self):
        # Inherited from CBasedExpressionWriter
        a, b, c, d = self.abcd
        self.eq(Not(NotEqual(a, b)), '(!(a != b))')
        self.eq(And(Equal(a, b), NotEqual(c, d)), '((a == b) && (c != d))')
        self.eq(Or(More(d, c), MoreEqual(b, a)), '((d > c) || (b >= a))')
        self.eq(Or(Less(d, c), LessEqual(b, a)), '((d < c) || (b <= a))')
        self.eq(
            Not(Or(Equal(Number(1), Number(2)), Equal(Number(3), Number(4)))),
            '(!((1.0 == 2.0) || (3.0 == 4.0)))')
        self.eq(Not(Less(Number(1), Number(2))), '(!(1.0 < 2.0))')

    def test_conditionals(self):
        # Inherited from CBasedExpressionWriter
        a, b, c, d = self.abcd
        self.eq(If(Equal(a, b), d, c), '((a == b) ? d : c)')
        self.eq(Piecewise(NotEqual(d, c), b, a), '((d != c) ? b : a)')
        self.eq(Piecewise(Equal(a, b), c, Equal(a, d), Number(3), Number(4)),
                '((a == b) ? c : ((a == d) ? 3.0 : 4.0))')


if __name__ == '__main__':
    unittest.main()
